(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: nsyntaxerr.ml,v 1.1.1.1 2002/01/16 09:01:22 cookcu Exp $ *)

(* Auxiliary type for reporting syntax errors *)

open Format
open Location

type err_kinds = IllegalAsPat | IllegalLabel | IllegalField | IllegalPat
               | IllegalRecBind | UnknownError | IllegalExpSeq
			   | IllegalType | IllegalTypeId | IllegalExp | IllegalVarId
			   | IllegalTypeVar | IllegalConId | IllegalDec | IllegalStrId
			   | IllegalFctId | IllegalSigId | IllegalStrExp | IllegalSigExp

type errmsg_kinds = EngGentle | HanGentle | HanTough

let errmsg_type = ref HanGentle

type error =
    Unclosed of Location.t * string * Location.t * string
  | Duptyid of Location.t * Location.t * string
  | Dupdatacon of Location.t * Location.t * string
  | Duppatvar of Location.t * Location.t * string
  | Orpaterr of Location.t
  | Other of Location.t * err_kinds

exception Error of error
exception Escape_error

let msglist = 
  [(IllegalAsPat,
    "As pattern must be used with variable pattern.",
    "As Ͽ ׻  ϰ  մϴ.",
    "As տ ׻  ´ٴϱ!!");
   (IllegalLabel,
    "Illegal record label.",
    "߸ ڵ ǥ(label)Դϴ.",
    "̻Ѱ ڲ ڵ ǥ , ڸ̴ ڸ̴ ؾ Ѵٴϱ!!");
   (IllegalField,
    "Illegal field of record pattern.",
    "ڵ ʵ尡 ̻մϴ.",
    "ڵ ȿٰ ̻Ѱ ??");
   (IllegalPat,
    "Illegal pattern.",
    "߸ Դϴ.",
    "̰͵ ̶ ??");
   (IllegalRecBind,
    "Recursive variable must be bound to 'fn' expression.",
    " ǵ  ׻ 'fn' ǥ ; մϴ.",
    "let rec  ҷ, 'fn' Ŷ, ٸ     𸣴ϱ.");
   (IllegalExpSeq,
    "Sequence of expression must end with expression(Not semi-colon).",
    "α׷ ǥ  ׻ α׷ ǥ  մϴ(ݷ ȵ˴ϴ).",
    "α׷ ǥ ҷ   α׷ ǥ̾ Ѵٱ. ݷ  ɷ  ȵ.");
   (IllegalType,
    "This should be a proper type.",
	"ùٸ Ÿ ; մϴ.",
	"ùٸ Ÿ ; մϴ.");
   (IllegalTypeId,
    "This should be proper type identifier.",
	"ùٸ Ÿ ĺڰ ; մϴ.",
	"ùٸ Ÿ ĺڰ ; մϴ.");
   (IllegalExp,
    "This should be a proper expression.",
	"ùٸ ǥ ; մϴ.",
	"ùٸ ǥ ; մϴ.");
   (IllegalVarId,
    "This should be proper variable identifier.",
	"ùٸ  ĺڰ ; մϴ.",
	"ùٸ  ĺڰ ; մϴ.");
   (IllegalTypeVar,
    "This should be proper type variable.",
	"ùٸ Ÿ  ; մϴ.",
	"ùٸ Ÿ  ; մϴ.");
   (IllegalConId,
    "This should be proper constructor identifier.",
	"ùٸ  ĺڰ ; մϴ.",
	"ùٸ  ĺڰ ; մϴ.");
   (IllegalDec,
    "This should be proper declaration.",
	"ùٸ  ; մϴ.",
	"ùٸ  ; մϴ.");
   (IllegalStrId,
    "This should be proper structure identifer.",
	"ùٸ  ĺڰ ; մϴ.",
	"ùٸ  ĺڰ ; մϴ.");
   (IllegalFctId,
    "This should be proper functor identifer.",
	"ùٸ  ĺڰ ; մϴ.",
	"ùٸ  ĺڰ ; մϴ.");
   (IllegalSigId,
    "This should be proper signature identifier.",
	"ùٸ ñ׳ó ĺڰ ; մϴ.",
	"ùٸ ñ׳ó ĺڰ ; մϴ.");
   (IllegalStrExp,
    "This should be proper structure expressions.",
	"ùٸ  ǥ ; մϴ.",
	"ùٸ  ǥ ; մϴ.");
   (IllegalSigExp,
    "This should be proper signature expressions.",
	"ùٸ ñ׳ó ǥ ; մϴ.",
	"ùٸ ñ׳ó ǥ ; մϴ.")
   ]
  
let print_err_kinds e =
  let (k,msg1,msg2,msg3) =
        try
          List.find (function (k,_,_,_) -> k = e) msglist
        with
          Not_found ->
	        (UnknownError,
             "Unknown error ocurred.",
             "ϴ  ߻߽ϴ.",
             "̰  𸣰ڴ.  ˾Ƽ ˾ƺ.")
   in print_string 
        (match !errmsg_type with
		   EngGentle -> msg1
		 | HanGentle -> msg2
		 | HanTough -> msg3)

let report_error = function
    Unclosed(opening_loc, opening, closing_loc, closing) ->
      if String.length !Location.input_name = 0
      && Location.highlight_locations opening_loc closing_loc
      then begin
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "Syntax error: '"
	   | HanGentle -> " : '"
	   | HanTough  -> "û Ǽ: '");
        print_string closing;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "' expected, the highlighted '"
	   | HanGentle -> "' ()  ֽϴ. ǥõ '"
	   | HanTough  -> "'  Գ!  ǥ '");
        print_string opening;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "' might be unmatched"
	   | HanGentle -> "' () ¦  ʴ  ϴ."
	   | HanTough  -> "'  ȸ ̴. Ķ.");
      end else begin
        Location.print std_formatter closing_loc;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "Syntax error: '"
	   | HanGentle -> " : '"
	   | HanTough  -> "û Ǽ: '");
        print_string closing;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "' expected "
	   | HanGentle -> "' ()  ֽϴ. "
	   | HanTough  -> "'  Ծ! "); 
	force_newline();
        Location.print std_formatter opening_loc;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "This '"
	   | HanGentle -> " ġ '"
	   | HanTough  -> " ǥ '");
        print_string opening;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "' might be unmatched"
	   | HanGentle -> "' () ¦  ʴ  ϴ."
	   | HanTough  -> "'  ȸ ̴. Ķ.")
      end
  | Duptyid(previous_loc, current_loc, name) ->
      if String.length !Location.input_name = 0
      && Location.highlight_locations previous_loc current_loc
      then begin
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "Syntax error: type name '"
	   | HanGentle -> " : Ÿ ̸ '"
	   | HanTough  -> "û Ǽ: Ÿ ̸ '");
        print_string name;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "' is duplicated at the highlighted positions"
	   | HanGentle -> "' () ϳ Ÿǳ ι Ͽϴ. "
			  ^"ǥõ κ   ǵ κ Դϴ."
	   | HanTough  -> "'  ǥ κп Ȱ  . ׷ ȵǿ~")
      end else begin
        Location.print std_formatter current_loc;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "Syntax error: type name '"
	   | HanGentle -> " : Ÿ ̸ '"
	   | HanTough  -> "û Ǽ: Ÿ ̸ '");
        print_string name;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "' previously appeared."; 
	   | HanGentle -> "'  ̹ ϳ Ÿǳ ̹ ǵǾϴ."
	   | HanTough  -> "'  ̹ տ ֱ. ׷ ȵǿ~~");
	force_newline();
        Location.print std_formatter previous_loc;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "Previous position of the type name '";
	   | HanGentle -> " ġ ̹ ǵǾ ֽϴ."
	   | HanTough  -> "   ,  ϳ ٲ.");
	if !errmsg_type = EngGentle
	then ( print_string name; print_string "'" )
	else ()
      end
  | Duppatvar(previous_loc, current_loc, name) ->
      if String.length !Location.input_name = 0
      && Location.highlight_locations previous_loc current_loc
      then begin
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "Syntax error: pattern variable '"
	   | HanGentle -> " :   '"
	   | HanTough  -> "û Ǽ:   '");
        print_string name;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "' is duplicated at the highlighted positions"
	   | HanGentle -> "' ()  Ͽ ι ϴ. "^
	   		  "ǥõ  Ȯϼ"
	   | HanTough  -> "'   Ͽ ι    ϶? "^
			  " ǥ  Ȯ .")
      end else begin
        Location.print std_formatter current_loc;
	print_string
	  (match !errmsg_type with
	     EngGentle -> "Syntax error: pattern variable '"
	   | HanGentle -> " :   '"
	   | HanTough  -> "û Ǽ:   '");
        print_string name;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "' previously appeared."
	   | HanGentle -> "' ()  ϳ ̹ ϴ."
	   | HanTough  -> "'  ̹  ִ ̱. ");
	force_newline();
	Location.print std_formatter previous_loc;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "Previous position of the pattern variable '"
	   | HanGentle -> " ġ ̹ ǵǾ ֽϴ."
	   | HanTough  -> "   ,  ϳ ٲ.");
	if !errmsg_type = EngGentle
	then ( print_string name; print_string "'" )
	else ()
      end
  | Dupdatacon(previous_loc, current_loc, name) ->
      if String.length !Location.input_name = 0
      && Location.highlight_locations previous_loc current_loc
      then begin
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "Syntax error: data constructor '"
	   | HanGentle -> " : Ÿ  '"
	   | HanTough  -> "û Ǽ: Ÿ  '");
        print_string name;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "' is duplicated at the highlighted positions."
	   | HanGentle -> "'   ǳ  ̻ ϴ."
	   | HanTough  -> "'  ι    ϳ? ϳ ٲ.")
      end else begin
        Location.print std_formatter current_loc;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "Syntax error: data constructor '"
	   | HanGentle -> " : Ÿ  '"
	   | HanTough  -> "û Ǽ: Ÿ  '");
        print_string name;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "' previously appeared."
	   | HanGentle -> "'   ǳ ̹ ̰ ֽϴ."
	   | HanTough  -> "'  ̹ ǵ ε? -.-");
	force_newline();
        Location.print std_formatter previous_loc;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "Previous position of the data constructor '"
	   | HanGentle -> " ġ ̹ ǵǾ ֽϴ."
	   | HanTough  -> "   ,  ϳ ٲ.");
        print_string name;
	print_string "'"
      end
  | Orpaterr (loc) ->
      if String.length !Location.input_name = 0
      && Location.highlight_locations loc loc
      then begin
      print_string 
	  (match !errmsg_type with
	     EngGentle -> "Syntax error: Or patterns must have same "^
	     		  "pattern variables"
	   | HanGentle -> " : Or   ϵ  "^
	   		  " ƾ մϴ."
	   | HanTough  -> "û Ǽ: Or ϵ鿡    "^
	   		  "ƾ Ѵٱ. ȱ׷ ̻ݾ?");
      end else begin
      Location.print std_formatter loc;
      print_string 
	  (match !errmsg_type with
	     EngGentle -> "Syntax error: Or patterns must have same "^
	     		  "pattern variables"
	   | HanGentle -> " : Or   ϵ  "^
	   		  " ƾ մϴ."
	   | HanTough  -> "û Ǽ: Or ϵ鿡    "^
	   		  "ƾ Ѵٱ. ȱ׷ ̻ݾ?");
      end
  | Other (loc,e) ->
      if String.length !Location.input_name = 0
      && Location.highlight_locations loc loc
      then begin
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "Syntax error: "
	   | HanGentle -> " : "
	   | HanTough  -> "û Ǽ: ");
        print_err_kinds e
      end else begin
        Location.print std_formatter loc;
        print_string 
	  (match !errmsg_type with
	     EngGentle -> "Syntax error: "
	   | HanGentle -> " : "
	   | HanTough  -> "û Ǽ: ");
        print_err_kinds e
      end

