(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: nsyntaxerr.n,v 1.2 2002/03/18 06:58:36 cookcu Exp $ *)

(* Auxiliary type for reporting syntax errors *)

signature SyntaxErrSig =
sig

type duplicate_type =
   DupTypeId | DupConId | DupVarId | DupLabel
 | DupStrId | DupFctId | DupSigId

type invalid_type =
   InvTypeId | InvTypeVar | InvConId | InvVarId | InvLabel
 | InvStrId | InvFctId | InvSigId
 | InvType | InvPat | InvAsPat | InvOrPat | InvExp | InvArrUpdate
 | InvDec | InvStrDec | InvRecBind | InvStrExp | InvSigExp
 | InvSpec | InvTopDec | InvTopDir

type msg_style = EngGentle | HanGentle | HanTough

type unmatched_type = Unclosed | DiffFunId

type error_kinds =
    Unmatched of unmatched_type * Location.t * string * Location.t * string
  | Duplicate of duplicate_type * Location.t * Location.t  * string
  | Invalid of invalid_type * Location.t
  | Predefined of Location.t * string
  | Missing of Location.t * string
  | Unknown of Location.t
  | Debug of Location.t * string
  (*
  | Orpaterr of Location.t
  *)

exception Error of error_kinds
exception Escape_error

val curmsg_style: msg_style ref
val report_error: error_kinds -> unit
end

structure NsyntaxErr : SyntaxErrSig = 
struct

open Format
open Location

type duplicate_type =
   DupTypeId | DupConId | DupVarId | DupLabel
 | DupStrId | DupFctId | DupSigId

type invalid_type =
   InvTypeId | InvTypeVar | InvConId | InvVarId | InvLabel
 | InvStrId | InvFctId | InvSigId
 | InvType | InvPat | InvAsPat | InvOrPat | InvExp | InvArrUpdate
 | InvDec | InvStrDec | InvRecBind | InvStrExp | InvSigExp
 | InvSpec | InvTopDec | InvTopDir

type msg_style = EngGentle | HanGentle | HanTough

type unmatched_type = Unclosed | DiffFunId

type error_kinds =
    Unmatched of unmatched_type * Location.t * string * Location.t * string
  | Duplicate of duplicate_type * Location.t * Location.t  * string
  | Invalid of invalid_type * Location.t
  | Predefined of Location.t * string
  | Missing of Location.t * string
  | Unknown of Location.t
  | Debug of Location.t * string
  (*
  | Orpaterr of Location.t
  *)

exception Error of error_kinds
exception Escape_error

val curmsg_style = ref HanGentle

fun print_loc x = Location.print Format.std_formatter x

fun duplicate_type_to_string x =
  let val (msg1, msg2, msg3) =
    case x of
      DupTypeId => ("type name", "Ÿ ̸", "Ÿ ̸")
    | DupConId => ("constructor name", " ̸", " ̸")
    | DupVarId => ("variable name", " ̸", " ̸")
    | DupLabel => ("record label", "ڵ ǥ", "ڵ ǥ")
    | DupStrId => ("structure name", " ̸", " ̸")
    | DupFctId => ("functor name", " ̸", " ̸")
    | DupSigId => ("signature name", "ñ׳ó ̸", "ñ׳ó ̸")
  in 
    case !curmsg_style of
      EngGentle => msg1
    | HanGentle => msg2
    | HanTough  => msg3
  end

fun invalid_type_to_string x =
  let val (msg1, msg2, msg3) =
    case x of
      InvTypeId => ("type name", "Ÿ ̸", "Ÿ ̸")
    | InvTypeVar => ("type variable", "Ÿ ", "Ÿ ")
    | InvConId => ("constructor name", " ̸", " ̸")
    | InvVarId => ("variable name", " ̸", " ̸")
    | InvLabel => ("record label", "ڵ ǥ", "ڵ ǥ")
    | InvStrId => ("structure name", " ̸", " ̸")
    | InvFctId => ("functor name", " ̸", " ̸")
    | InvSigId => ("signature name", "ñ׳ó ̸", "ñ׳ó ̸")
    | InvType => ("type", "Ÿ", "Ÿ")
    | InvAsPat => ("as pattern", "as ", "as ")
    | InvOrPat => ("or pattern", "or ", "or ")
    | InvPat => ("pattern", "", "")
    | InvExp => ("expression", "α׷ ǥ", "α׷ ǥ")
    | InvArrUpdate => ("array update", "迭 ", "迭 ")
    | InvDec => ("declaration", "α׷ ", "α׷ ")
    | InvStrDec => ("structure declaration", " ", " ")
    | InvRecBind => ("recursive binding", " ", " ")
    | InvStrExp => ("structure expression", " ǥ", " ǥ")
    | InvSigExp => ("signature expression", "ñ׳ó ǥ", "ñ׳ó ǥ")
    | InvSpec => ("specification", "ӹ", "ӹ")
    | InvTopDec => ("toplevel declaration", "ֻ ", "ֻ ")
    | InvTopDir => ("toplevel directive", "ֻ ", "ֻ ")
  in 
    case !curmsg_style of
      EngGentle => msg1
    | HanGentle => msg2
    | HanTough  => msg3
  end


fun print_head () = 
  (case !curmsg_style of
     EngGentle => print_string "Syntax error: "
   | HanGentle => print_string ": "
   | HanTough  => print_string "û Ǽ: ")

fun print_msg (enggentle, hangentle, hantough) = 
  (case !curmsg_style of
     EngGentle => print_string enggentle
   | HanGentle => print_string hangentle
   | HanTough  => print_string hantough)
  
fun report_error(Unmatched(t, first_loc, first, second_loc, second)) =
      ((if not (String.length !Location.input_name = 0)
         || not (Location.highlight_locations first_loc second_loc)
      then (print_loc first_loc; print_loc second_loc));
      print_head();
      print_msg(second^" is not matched with "^first^".",
        second^"() "^first^"  ʽϴ.",
        second^" "^first^"()  ʴµ."))
  | report_error(Duplicate(t, first_loc, second_loc, id)) =
      ((if not (String.length !Location.input_name = 0)
         || not (Location.highlight_locations first_loc second_loc)
      then (print_loc first_loc; print_loc second_loc));
      print_head();
      let val msg = duplicate_type_to_string t
       in print_msg(msg^" "^id^" are duplicated.",
            "ߺ "^msg^" "^id^"() Ǿϴ.",
            msg^" "^id^"()  ġµ.")
      end)
  | report_error(Invalid(t, loc)) =
      ((if not (String.length !Location.input_name = 0)
         || not (Location.highlight_locations loc loc)
      then print_loc loc);
      print_head();
      let val msg = invalid_type_to_string t
       in print_msg("invalid "^msg^" is used.",
            "߸ "^msg^"() Ǿϴ.",
            msg^"() ̻.")
      end)
  | report_error(Predefined(loc, id)) =
      ((if not (String.length !Location.input_name = 0)
         || not (Location.highlight_locations loc loc)
      then print_loc loc);
      print_head();
      print_msg("reserved name "^id^" is used.",
        "̸ ǵ ̸ "^id^"() Ǿϴ.",
        id^"() ̹ صξ  ."))
  | report_error(Missing(loc, id)) =
      ((if not (String.length !Location.input_name = 0)
         || not (Location.highlight_locations loc loc)
      then print_loc loc);
      print_head();
      print_msg("'"^id^"' is missing.",
        "'"^id^"'() ϴ.",
        "'"^id^"'() ̹ صξ  ."))
  | report_error(Unknown(loc)) =
      ((if not (String.length !Location.input_name = 0)
         || not (Location.highlight_locations loc loc)
      then print_loc loc);
      print_head();
      print_msg("unknown error.",
        "˼  .",
        "𸣴 ε."))
  | report_error(Debug(loc,s)) =
      ((if not (String.length !Location.input_name = 0)
         || not (Location.highlight_locations loc loc)
      then print_loc loc);
      print_head();
      print_msg("Debug:"^s,
        "Debug:"^s,
        "Debug:"^s))
end
