(*
 Oukseh Lee, The LET Project, KAIST

 Copyright(c) 2000-2002 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

open Format 

structure Infer = struct

  open Ty
  open NEnv
  open String_ast.Ast

  structure I = Absyn
  structure R = TyError

  val foldr = List.fold_right

  val b2d = fn (TypeBind(a,b,c,d)) => (TypeBindDesc(a,b,c,d))
             | (DataBind(a,b,c,d)) => (DataDesc(a,b,c,d))
  val d2b = fn (I.TypeBindDesc(a,b,c,d)) => (I.TypeBind(a,b,c,d))
             | (I.DataDesc(a,b,c,d)) => (I.DataBind(a,b,c,d))
             | _ => raise TyDebug.Bug "Infer.d2b"


  fun is_expansive vbl = 
    let fun f e = case e of
            I.UnitExp _ | I.IntExp _ | I.RealExp _ | I.StringExp _ | I.CharExp _
          | I.VarExp _ | I.FnExp _ | I.ConExp _ => false
          | I.RecordExp(ll,_,_) => List.exists (fn (_,e,_)=> f e) ll
          | I.TupleExp(el,_,_)
          | I.ListExp(el,_,_) => List.exists f el
          | I.AppExp(e',e,_,_) => g e' || f e
          | I.ConstraintExp(e,_,_) | I.SeqExp([e],_,_) => f e
          | _ => true
        and g e = case e of
            I.ConExp _ => false
          | I.ConstraintExp(e,_,_) => g e
          | _ => true
    in  List.exists (fn (_,_,e,_) => f e) vbl 
    end

  fun exp_loc exp = case exp of 
      UnitExp i | IntExp (_,i) | RealExp (_,i) | StringExp (_,i) | CharExp (_,i)
    | VarExp (_,i) | ConExp (_,i) | AppExp (_,_,i) | RecordExp (_,i) 
    | RecordFieldExp (_,_,i) | ArrayFieldExp (_,_,i) | UpdateArrayExp (_,_,_,i)
    | SubstRecordExp (_,_,_,i) | TupleExp (_,i) | ListExp (_,i) | ArrayExp (_,i)
    | LetExp (_,_,i) | HandleExp (_,_,i) | RaiseExp (_,i) | FnExp (_,i)
    | AssignExp (_,_,i) | RefExp (_,i) | DeRefExp (_,i) | SeqExp (_,i)
    | CaseExp (_,_,i) | IfExp (_,_,_,i) | WhileExp (_,_,i) | ForExp (_,_,_,_,_,i) 
    | ConstraintExp(_,_,i) => i


  fun make_app e1 e2 =
    AppExp(e1, [e2], Location.location_to_location
                     (Location.location_field_loc_start (exp_loc e1),
                      Location.location_field_loc_end (exp_loc e2),
                      false))

  fun pairing l = case l of
      [] => []
    | (e1 as ConExp(v,_))::e2::t => 
        let val ((_,ty),_) = find_ty v
        in  if Ty.is_fun ty then (make_app e1 e2) :: (pairing t) 
                            else e1 :: (pairing (e2::t))
        end
    | e1::t => e1::(pairing t)

  fun resolve_app l = case l of
      [] => raise TyDebug.Bug "make_app"
    | [h] => h
    | h::t => make_app (resolve_app t) h

  fun expr_ty e rho = let 
   fun f e rho = case e of

  (* Rule 4.65 *)
      UnitExp a =>      (Unify.const a rho type_unit;   I.UnitExp a)
    | IntExp(v,a) =>    (Unify.const a rho type_int;    I.IntExp(v,a))
    | RealExp(v,a) =>   (Unify.const a rho type_real;   I.RealExp(v,a))
    | StringExp(v,a) => (Unify.const a rho type_string; I.StringExp(v,a))
    | CharExp(v,a) =>   (Unify.const a rho type_char;   I.CharExp(v,a))

  (* Rule 4.66 *)
    | VarExp(v,a) => (let val (t,id) = find_ty v 
                          val (_,(s,_),_) = v
                      in Unify.it a rho (Ty.instance a t); 
                         I.VarExp(id,rho,a) 
                      end)
   (*       handle R.T(y as (R.Unbound _,_)) => (  (* option !! *)
            R.warning y; 
            I.VarExp(v,rho,a))) *)

    | ConExp(v,a) => (let val (t,id) = find_ty v in
        Unify.it a rho (Ty.instance a t); I.ConExp(id,rho,a) end)
(*          handle R.T(y as (R.Unbound _,_)) => (  (* option !! *)
            R.warning y;
            I.ConExp(v,rho,a))) *)
   
  (* Rule 4.67 *)   
  (* Rule 4.84 *)
    | RecordExp(r,a) => let   
        val m = ref Labm.M.empty
        fun g ((lab,a),e,b) = let val beta = Tv.new() 
                              in  m := Labm.M.add lab beta (!m);
                                  ((lab,a),f e beta,b) 
                              end
        val r' = List.map g r
       in
        Unify.it a rho (Row.closerow(!m));
        I.RecordExp(r',rho,a)
       end
      
  (* Rule 4.68 *)   
    | RecordFieldExp(e,(l,a),a') => 
       let val tau = Row.openrow_from_olist [(l,rho)]
       in  I.RecordFieldExp(f e tau,(l,a),rho,a') 
       end

  (* Rule 4.69 *)
    | ArrayExp(el,a) => let
        val tau = Tv.new()
       in
        Unify.data a rho (gen_array tau);
        I.ArrayExp(List.map (fn e=>f e tau) el,rho,a)
       end

  (* Rule 4.70 *)
    | ArrayFieldExp(e1,e2,a) => let 
        val e1' = f e1 (gen_array rho) 
        val e2' = f e2 type_int
       in
        I.ArrayFieldExp(e1',e2',rho,a)
       end

  (* Rule 4.71 *)
    | UpdateArrayExp(e1,e2,e3,a) => let
        val tau = Tv.new()
        val _ = Unify.it a rho type_unit
        val e1' = f e1 (gen_array tau)
        val e2' = f e2 type_int
        val e3' = f e3 tau
       in
        I.UpdateArrayExp(e1',e2',e3',rho,a)
       end
         
  (* Rule 4.72 *)
    | LetExp(d,e,a) => 
       let val _ = push_e (E.empty())
           val d' = dec_env d
           val e' = f e rho
       in  ignore(pop());
           I.LetExp(d',e',rho,a)
       end

    | AppExp(_,[],_) => raise TyDebug.Bug "AppExp(_,[],_)"

  (* Rule 4.74 in OCaml sytle *)
    | AppExp(e1,[e2],a) => let
        val tau = Tv.new() val tau1 = Tv.new()  (* OCaml Style *)
        val e1' = f e1 tau
        val _ = Unify.app a (Absyn.expLoc e1') tau tau1 rho
        val e2' = f e2 tau1
       in
        I.AppExp(e1',e2',rho,a)
       end

  (* Rule 4.74 in H style 

    | AppExp(e1,e2,_) => let
        val (beta1,beta2) = (Tv.new(),Tv.new());
       in
        expr_type' e1 (gen_fun(beta1,beta2));
        Unify.it[beta2,rho];
        expr_type' e2 beta1
       end
  *)
    | AppExp(e,l,a) => f (resolve_app (List.rev (pairing (e::l)))) rho
    
  (* Rule 4.75 *)

    | ConstraintExp(e,ty,a) => (ty_ty ty rho; I.ConstraintExp(f e rho, ty, a))
  
  (* Rule 4.76 *)
    | FnExp(m,a) => I.FnExp(fnrule_ty m rho,rho,a)

  (* Rule 4.77 *)
    | HandleExp(e,m,a) => let
        val e' = f e rho 
        val m' = match_ty m (gen_fun type_exn rho)
       in
        I.HandleExp(e',m',rho,a)
       end

  (* Rule 4.78 *)
    | RaiseExp(e,a) => I.RaiseExp(f e type_exn,rho,a)

  (* Rule 4.79 *)
    | RefExp(e,a) => let
        val tau = Tv.new()
       in
        Unify.it a rho (gen_ref tau);
        I.RefExp(f e tau,rho,a)
       end

  (* Rule 4.80 *)
    | AssignExp(e1,e2,a) => let
        val tau = Tv.new()
        val _ = Unify.it a rho type_unit
        val e1' = f e1 (gen_ref tau)
        val e2' = f e2 tau
       in
        I.AssignExp(e1',e2',rho,a)
       end
   
  (* Rule 4.81 *)
    | DeRefExp(e,a) => I.DeRefExp(f e (gen_ref rho),rho,a)

  (* Rule 4.82 *)
    | SeqExp([],_) => raise TyDebug.Bug "Infer.expr_ty.SeqExp.Null seq of exp"
    | SeqExp([e],_) => f e rho
    | SeqExp(el,a) => let
        val tl = (list_init ((List.length el)-1) (fn _ => Tv.new()))@[rho]
       in
        I.SeqExp(List.map2 f el tl,rho,a)
       end

  (* Rule 4.83 *)
    | CaseExp(e,m,a) => let
        val tau = Tv.new()
        val e' = f e tau
        val m' = match_ty m (gen_fun tau rho)
       in
        I.CaseExp(e',m',rho,a)
       end

  (* SubstRecordExp *)
    | SubstRecordExp(e1,(lab,a),e2,b) => let
        val beta = Tv.new()
        val _ = Unify.it a rho (Row.openrow_from_olist [(lab,beta)])
        val e1' = f e1 rho
        val e2' = f e2 beta
       in
        I.SubstRecordExp(e1',(lab,a),e2',rho,b)
       end

  (* TupleExp *) 
    | TupleExp(el,a) => let
        val tl = list_init (List.length el) (fn _ => Tv.new())
        val _ = Unify.tuple a rho tl
        val el = List.map2 f el tl
       in
        I.TupleExp(el,rho,a)
       end
                         
  (* ListExp extended from Rule 4.67, 4.74 *)
    | ListExp(el,a) => let
        val beta = Tv.new()
       in
        Unify.data a rho (gen_list beta);
        I.ListExp(List.map (fn e => f e beta) el,rho,a)
       end

  (* IfExp extended from Rule 4.83 *)
    | IfExp(e1,e2,e3,a) => 
          let val e1' = f e1 type_bool
              val _ = if e3 = UnitExp(Location.none) then Unify.it a type_unit rho
              val e2' = f e2 rho
              val e3' = f e3 rho
          in  I.IfExp(e1',e2',e3',rho,a)
          end

  (* WhileExp *)
    | WhileExp(e1,e2,a) => let
        val _ = Unify.it a rho type_unit
        val e1' = f e1 type_bool
        val e2' = f e2 (Tv.new())
       in
        I.WhileExp(e1',e2',rho,a)
       end

  (* ForExp *)
    | ForExp(x,e1,e2,e3,e4,a) => let
        val beta1 = Tv.new() val beta2 = Tv.new()
        val _ = Unify.it a rho type_unit
        val e1' = f e1 beta1
        val _ = push_ve (VE.new x (in_scheme beta1))
        val e2' = f e2 type_bool
        val e3' = f e3 beta1
        val e4' = f e4 beta2
        val _ = pop()
       in
        I.ForExp(x,e1',e2',e3',e4',rho,a)
       end
   in f e rho
   end
  
  (* Rule 4.85 and 4.86 *)
  and match_ty nil _ = raise TyDebug.Bug "infer.match_ty.NullMatch"
    | match_ty rl rho = let 
        val tau = Tv.new() val tau' = Tv.new()
        val _ = (Unify.lambda ((**) Location.none) rho tau tau')
                   handle _ => raise TyDebug.Bug "infer.match_ty.Unify.lambda"
        fun f (p,e,b) = 
          let val ve = VE.empty()
              val _ = push_ve ve
              val p' = pat_ve p tau ve
              val e' = expr_ty e tau'
              val _ = pop()
          in  (p',e',b) end
       in List.map f rl end 

  and fnrule_ty nil _ = raise TyDebug.Bug "infer.fnrule_ty.NullMatch"
    | fnrule_ty rl rho = let
        val len = ref None
        fun f (pl,e,a) = let
          val ve = VE.empty()
          val _ = push_ve ve
          fun g rho [] = ([],rho)
            | g rho (ConPat(cid,a)::p::pl) =
                let val (t,id) = find_ty cid
                    val t' = Ty.instance a t
                in  case unlink t' of
                      Ty.Fun(t1,t2) => 
                        let val t3 = Tv.new()
                            val _ = Unify.it a rho (gen_fun t2 t3)
                            val p' = pat_ve p t1 ve
                            val (pl',rho') = g t3 pl
                        in  (I.AppPat(id,p',t2,a)::pl',rho')
                        end
                    | _ => 
                        let val t2 = Tv.new()
                            val _ = Unify.it a rho (gen_fun t' t2)
                            val (pl',rho') = g t2 (p::pl)
                        in  (I.ConPat(id,t',a)::pl',rho')
                        end
                end
            | g rho (p::pl) = 
                let val t1 = Tv.new() val t2 = Tv.new()
                    val _ = Unify.lambda a rho t1 t2
                    val p' = pat_ve p t1 ve
                    val (pl',rho') = g t2 pl
                in  (p'::pl',rho') 
                end
          val (pl',rho') = g rho pl
          val _ = case !len of None => len := Some (List.length pl')
                             | Some n => let val l = List.length pl'
                                         in  if n<>l then
                                             raise TyError.T(TyError.FunLength(n,l),a)
                                         end
          val e' = expr_ty e rho'
          val _ = pop()
         in (pl',e',a) end
       in List.map f rl end
 
  and pat_ve pat rho ve = let
   fun f rho pat = 
    case pat of

  (* Rule 4.87 *)
      WildPat a => I.WildPat(rho,a)
    
  (* Rule 4.88 *)
    | UnitPat a =>      (Unify.const a rho type_unit;   I.UnitPat a)
    | IntPat(v,a) =>    (Unify.const a rho type_int;    I.IntPat(v,a))
    | StringPat(v,a) => (Unify.const a rho type_string; I.StringPat(v,a))
    | CharPat(v,a) =>   (Unify.const a rho type_char;   I.CharPat(v,a))
   
  (* Rule 4.89 *)
    | RecordPat(patrow,a) => let
        val (patrow',row) = patrow_r patrow
       in
        Unify.it a rho (Row.closerow row);
        I.RecordPat(patrow',rho,a)
       end

  (* Rule 4.89 and 4.100 *)
    | SubRecordPat(patrow,a) => let
        val (patrow',row) = patrow_r patrow
       in
        Unify.it a rho (Row.openrow row);
        I.SubRecordPat(patrow',rho,a)
       end 

  (* Rule 4.90 *)
    | ArrayPat(pl,a) => let val tau = Tv.new() in 
        Unify.data a rho (gen_array tau);
        I.ArrayPat(List.map (f tau) pl,rho,a) 
       end

  (* Rule 4.92 *)
    | RefPat(pat,a) => let val tau = Tv.new() in
        Unify.data a rho (gen_ref tau);
        I.RefPat(f tau pat,rho,a)
       end

  (* Rule 4.93 *)
    | VarPat(x,b) => (VE.exadd ve x (in_scheme rho); I.VarPat(x,rho,b))
    
  (* Rule 4.95 *)
    | ConPat(cid,a) => let val (t,id) = find_ty cid in
        Unify.it a rho (Ty.instance a t); 
        I.ConPat(id,rho,a)
       end
    
  (* Rule 4.96 in OCaml style *)
    | AppPat(conid as (_,_,d),pat,a) => 
       let val tau = Tv.new() val tau' = Tv.new()
       in  case f tau (ConPat(conid,d)) of 
             I.ConPat(conid',_,b) => (
               Unify.app a b tau tau' rho;
               let val pat' = f tau' pat
               in I.AppPat(conid',pat',rho,a) end)
           | _ => raise TyDebug.Bug "Infer.pat_vepe.AppPat" 
       end

  (* Rule 4.97 *)

    | ConstraintPat(pat,ty,a) => (ty_ty ty rho; I.ConstraintPat(f rho pat, ty, a))

  (* Rule 4.98 *)
    | AsPat(x,typ,pat,b) => let
        val _ = case typ of Some ty => ty_ty ty rho | _ => ()
        val pat' = f rho pat
       in
        VE.exadd ve x (in_scheme rho); 
        I.AsPat(x,pat',rho,b)
       end

  (* Rule 4.99 *)
    | OrPat(p::pl,a) => let
        val vep = VE.empty()
        val p' = pat_ve p rho vep
        fun g p = let 
          val vep' = VE.empty()
          val p' = pat_ve p rho vep'
         in unify_ve a vep vep'; p' end
       in
        VE.explus ve vep;
        I.OrPat(p'::(List.map g pl),rho,a)
       end
         
  (* TuplePat *)
    | TuplePat(pl,a) => let
        val tl =  list_init (List.length pl) (fn _ => Tv.new())
       in
        Unify.tuple a rho tl;
        I.TuplePat(List.map2 f tl pl,rho,a)
       end

  (* ListPat *)
    | ListPat(pl,a) => let
        val tau = Tv.new()
        val _ = Unify.data a rho (gen_list tau)
        val pl' = List.map (f tau) pl
       in
        I.ListPat(pl',rho,a)
       end
 
    | _ => raise TyDebug.Bug "Infer.pat_ve.OrPat([],_) occurred."

    and patrow_r labpat = let
      val (labpat', labl) = List.split(
             List.map (fn ((lab,a),p,b) => let
                             val beta = Tv.new() val p' = f beta p
                            in (((lab,a),p',b),(lab,beta)) end) labpat)
     in (labpat', Labm.of_list labl) end
   in f rho pat end 

  and ty_ty ty rho = let
    fun f rho ty = case ty of
      (VarTy(tv,a)) => Unify.it a rho (find_tv tv)
    | (ConstTy(tyl,tyid,a)) => let
        val tyl = List.map (fn ty=>let val t=Tv.new() in f t ty; t end) tyl
        val (tyfn,_) = find_tyfn tyid
       in
        (Unify.it a rho (TS.apply tyl tyfn))
           handle TS.Apply_error => R.ts_apply tyid (TS.arity tyfn) (List.length tyl) a
       end

    | (RecordTy(ltl,a)) => let (* W style *)
        val l = List.map (fn((l,_),ty,_)=>let val t=Tv.new() in f t ty;(l,t)end) ltl
       in
        Unify.it a rho (Row.closerow(Labm.of_list l))
       end

    | (TupleTy(tyl,a)) => let (* M style *)
        val tl = list_init (List.length tyl) (fn _ => Tv.new())
       in
        Unify.tuple a rho tl;
        List.iter2 f tl tyl
       end

    | (FunTy(ty1,ty2,a)) => let  (* M style *)
        val tau1 = Tv.new() val tau2 = Tv.new()
       in
        Unify.lambda a rho tau1 tau2;
        f tau1 ty1;
        f tau2 ty2
       end
   in f rho ty end

  and dec_env dec = let
    fun ff dec = case dec of

      ValDec(tvl,vbl,a) => let
        val c = Tv.current()
        val _ = push_tvs (tvl @ Scope.find a)
        val ve = VE.empty() and ver = VE.empty()
        val (n0,r0) = List.partition (fn (Nonrec,_,_,_) => true | _ => false) vbl
        val n1 = List.map (fn (_, p, e, a) => 
                            let val t = Tv.new()
                            in  (I.Nonrec, pat_ve p t ve, expr_ty e t, a)
                            end) n0
        val r1 = List.map (fn (_, p, e, a) =>
                            let val t = Tv.new()
                            in  (I.Rec, pat_ve p t ver, e, t, a)
                            end) r0
        val _ = push_ve ver
        val r2 = List.map (fn (b, p, e, t, a) => (b, p, expr_ty e t, a)) r1
        val _ = VE.explus ve ver and vbl' = n1 @ r2
        val ve' = if is_expansive vbl' then ve else clos_ve c ve
    in  pop_tvs(); ignore(pop());
        add_ve ve';
        I.ValDec(tvl, vbl', VE.to_list ve', a)
    end

    | FunDec(tvl,fbl,a) => let
        val _ = push_tvs (tvl @ Scope.find a)
        fun ffunbind (bl,a) =
          let val i = ref None
              fun fbody ((j,b),pl,e,c) =
                case !i of None => i := Some(j,b); (pl,e,c)
                         | Some (i',_) => if i'=j then (pl,e,c) 
                                                  else raise R.T(R.FunName(i',j),b)
              val bl' = List.map fbody bl
          in  case !i of None => raise TyDebug.Bug "infer.FunDec"
                       | Some i' => (Rec, VarPat(i',a), FnExp(bl',a),a)
          end
        val z = ff(ValDec(tvl, List.map ffunbind fbl, a))
     in pop_tvs();
        z
     end 
        
    | TypeDec(tbl,a) => I.TypeDec(List.map d2b (typbind (List.map b2d tbl) a),a)

    | ExceptionDec(ebl,a) => I.ExceptionDec(conbind_ve type_exn ebl,a)

    | LocalDec(d1,d2,a) => 
       let val _ = push_empty()
           val d1' = dec_env d1
           val _ = reduce()
           val _ = push_empty()
           val d2' = dec_env d2
           val e = pop()
       in  ignore(pop()); add_e e; I.LocalDec(d1',d2',a)
       end

    | SeqDec(dl,a) => I.SeqDec(List.map dec_env dl,a)

    | OpenDec(l,a) => I.OpenDec(add_open l,a)
(*
    (* structure ~A1 = A1 ... structure ~An = An
                         open ~A1 ... ~An *) 
        let val l' = 
        in  I.OpenDec(add_open l,a)
        end
*)

   in ff dec end

  and typbind tbl a = let
        fun split (l1,l2,l3) (TypeDesc(a,b,c)) = ((a,b,c)::l1,l2,l3)
          | split (l1,l2,l3) (TypeBindDesc(a,b,c,d)) = (l1,(a,b,c,d)::l2,l3)
          | split (l1,l2,l3) (DataDesc(a,b,c,d)) = (l1,l2,(a,b,c,d)::l3)
        val (tn,ts,ds) = List.fold_left split ([],[],[]) tbl

        (* add data types into TE *)
        val ds' = List.map (fn (tl,id,b,c) =>
                             let val ts = Tn.new ((fn(x,_)=>x) id) (List.length tl) 
                                          (!context)
                             in  add_tei id (TS.gen_name ts); (tl,(id,ts),b,c) end) ds
        val _ = List.iter (fn (tl,id,a) =>
                             let val ts = TS.gen_var id (List.length tl) 
                                          (!context)
                             in  add_tei id ts end) tn

        (* order typebinds and add into TE *)
        fun iter l tns = let
          fun f (x as (tl,id,ty,b),used) (s,l,r) =
           let val s' = Tids.M.inter tns used
           in  if s' = Tids.M.empty then
                let val tau = Tv.new()
                in  push_tvs_bound tl;
                    ty_ty ty tau;
                    pop_tvs();
                    add_tei id (TS.gen_fun (List.length tl) tau);
                    (Tids.add id s, I.TypeBindDesc(tl,id,(ty,tau),b)::l, r)
                end
               else (s,l,(x,s')::r)
           end
          val (s',l',r') = foldr f l (Tids.M.empty, [],[])
         in if r' = [] then l'
            else if s' <> Tids.M.empty then l'@(iter r' (Tids.M.diff tns s')) 
            else raise R.T(R.RecursiveType(
                             List.map (fn ((_,(id,_),_,_),_)=> id) r'),
                           a)
         end
        val ts' = iter (List.map (fn (x as (_,_,ty,_)) => (x,Ty.used_tyids ty)) ts)
                       (foldr (fn (_,id,_,_) => Tids.add id) ts Tids.M.empty)

        (* update constructors into VE *)
        fun f (tl,(id,tn),cl,b) = 
         let val _ = push_tvs_bound tl
             val tau = TS.apply (List.map find_tv tl) (TS.gen_name tn)
             val acl = conbind_ve tau cl
         in  pop_tvs();
             CE.add tn (List.map (fn((id,_),None,_) => (id,None)
                                   |((id,_),Some (_,t),_) => (id,Some t)) acl);
             I.DataDesc(tl,id,acl,b) end
        val ds'' = List.map f ds'
        val tn' = List.map (fn x => I.TypeDesc x) tn
       in tn'@ts'@ds'' end  

  and conbind_ve tau cl = 
    let fun f (id,None,a) = (add_vei id (Ty.clos 0 tau); (id,None,a))
          | f (id,Some t,a) = let val tau' = Tv.new()
                              in  ty_ty t tau';
                                  add_vei id (Ty.clos 0 (gen_fun tau' tau));
                                  (id,Some(t,tau'),a)
                              end
    in  List.map f cl end

  fun strdec sd = case sd of
      SimpleDec(d,a) => I.SimpleDec(dec_env d,a)
    | StrDec(sbl,a) =>
       let fun g (id,None,e,a) = (id,e,a)
             | g (id,Some s,e,a) = (id,SigStr(e,s,a),a)
           fun f sb = let val (id,e,a) = g sb
                          val _ = push_context ((fn(s,_)=>s) id)
                          val (e',env) = strexp_e e
                      in  pop_context();
                          (* add_sei id env; *)
                          ((id,None,e',a),env)
                      end
           fun h (x as (id,_,_,_), env) = (add_sei id env; x)
       in  I.StrDec(List.map h (List.map f sbl),a) end
    | SeqStrDec([sd],a) => strdec sd
    | SeqStrDec(sdl,a) => I.SeqStrDec(List.map strdec sdl,a)
(*
       let fun f sd = let val _ = push_empty()
                          val sd' = strdec sd
                      in  (sd', pop())
                      end
           fun g (sd,e) = (add_e e; sd)
       in  I.SeqStrDec(List.map g (List.map f sdl),a)
       end

        I.SeqStrDec(List.map strdec sdl,a) *)
  and strname se = case se of
      VarStr(id,_) => Some (R.to_lid id)
    | _ => None
  and strexp_e se = case se of 
      VarStr(id,a) =>  let val (e',id') = find_e id
                       in  (I.VarStr(id',a),e')
                       end
    | StrStr(e,a) =>   let val _ = push_empty()
                           val e' = strdec e
                       in  (I.StrStr(e',a), pop())
                       end 
    | SigStr(e,s,a) => let val (e',env) = strexp_e e
                           val (s',sigma) = sigexp_sig s
                       in  ((ignore(Match.eg env sigma))
                              handle R.M x => raise R.T(R.Match(R.SS,strname e,signame s,x),a)); 
                           (I.SigStr(e',s',a), (G.to_env sigma))
                       end
    | FctAppStr(id,sl,a) => 
        let val f = find_f id
            val str = case sl of
                  [] => raise TyDebug.Bug "FctAppStr(_,[],_)"
                | [h] => h
                | _ => F.make_argument f sl
            val old_context = !context
            val _ = reset_context()
            val (str',e) = strexp_e str
            val _ = set_context old_context
            val e' = (Match.ef e f (!context))
                       handle R.M x => raise R.T(R.Match(R.FI,strname str,fctname id,x),a)
        in  (I.FctAppStr(id,[str'],a),e')
        end

  and fctname (id,_) = Some [id]
  and spec sp = case sp of
      ValSpec(vdl,a) => 
        let fun f (id,ty,a) = let val _ = push_tvs (Ty.used_tyvars ty)
                                  val tau = Tv.new()
                              in  ty_ty ty tau;
                                  pop_tvs();
                                  add_vei id (Ty.clos 0 tau);
                                  (id,(ty,tau),a)
                              end
        in  I.ValSpec(List.map f vdl,a) 
        end
    | TypeSpec(tdl,a) => I.TypeSpec(typbind tdl a,a)
    | ExnSpec(edl,a) => I.ExnSpec(conbind_ve type_exn edl,a)
    | StrSpec(sdl,a) => 
        let fun f(id,e,a) = let val _ = push_context ((fn(x,_)=>x) id)
                                val (e',sigma) = sigexp_sig e
                                val _ = pop_context ()
                            in  add_sei id (G.to_env sigma);
                                (id,e',a)
                            end
        in  I.StrSpec(List.map f sdl,a)
        end
    | IncludeSpec(e,a) => let val (e',sigma) = sigexp_sig e
                          in  add_e (G.to_env sigma);
                              I.IncludeSpec(e',a)
                          end
    | SeqSpec([s],a) => spec s
    | SeqSpec(sl,a) => I.SeqSpec(List.map spec sl,a)

  and signame se = case se of
      VarSig((id,_),_) => Some [id]
    | _ => None
  and sigexp_sig se = case se of
      VarSig(id,a) => (I.VarSig(id,a), G.make_new (find_sig id) (!context))
    | SigSig(s,a) => 
       let val _ = push_empty()
           val c = Tn.current()
           val s' = spec s
           val sigma = G.from_env c (pop())
       in  (I.SigSig(s',a),sigma)
       end
    | ConstraintSig(s,(ltbl,b),a) => 
       let val (s',sigma) = sigexp_sig s
           val e = G.to_env sigma
           val sub = ref TSubst.empty
           val trm = ref (G.t sigma)
           fun f (tl,id,ty,a) = case find_tyfn_in e id of
               TS.TsFun _ => R.constraint_sig id
             | TS.TsName tn => 
                if CE.find tn <> [] then R.constraint_sig id
                else if TSubst.mem tn (!sub) then R.redundant_constraint_sig id 
                else
                  let val _ = push_tvs_bound tl
                      val tau = Tv.new()
                  in  ty_ty ty tau;
                      pop_tvs();
                      sub := TSubst.add tn (TS.gen_fun (List.length tl) tau) (!sub);
                      trm := Tns.remove tn (!trm);
                      (tl,id,(ty,tau),a)
                  end
           val ltbl' = List.map f ltbl
       in  (I.ConstraintSig(s',(ltbl',b),a), (!trm, TSubst.apply_e (!sub) e))
       end

  and sigdec (sl,a) = 
    let fun f (id as (ids,_),e,a) = let val _ = push_context (sig_context' ids) 
                             val (e',sigma) = sigexp_sig e
                         in  pop_context(); 
                             add_sig id sigma;
                             (id,e',a)
                         end
    in  (List.map f sl,a)
    end

  and fctdec (id,fal,sigopt,se,a) = 
    let val sw = ref None
        val (idl,(sid,sg,b)) = 
              case fal of 
                [(s,e,a)] => ([s],(s,e,a)) 
              | _ => let val sid = (functor_context(),a)
                     in  sw := Some sid;
                         (List.map (fn (x,_,_)=>x) fal,
                          (sid,SigSig(StrSpec(fal,a),a),a))
                     end
        val se = case sigopt of None => se | Some si => SigStr(se,si,a)
        val _ = push_context ((fn(x,_)=>x) sid)
        val (sg',sigma) = sigexp_sig sg
        val _ = (pop_context (); push_context (fct_context id))
        val _ = push_e (E.empty())
        val _ = add_sei sid (G.to_env sigma)
        val c = Tn.current()
        val se = case (!sw) of 
                   Some s => 
                     let fun g sd = SeqStrDec([SimpleDec(OpenDec([([],s,a)],a),a),sd],a)
                         fun h (x as VarStr _) = x
                           | h (StrStr(sd,b)) = StrStr(g sd,b)
                           | h (SigStr(e,g,b)) = SigStr(h e,g,b)
                           | h (FctAppStr(i,el,b)) = FctAppStr(i,List.map h el,b)
                     in h se
                     end
                 | None => se
(*        val _ = (* DEBUG *) EnvPrint.b (!(NEnv.s2));
                            EnvPrint.estack (!(NEnv.s3))
*)
        val (se',e') = (strexp_e se)  handle exc as R.T(R.Match(_,_,s,m),b) =>
                                       if a=b then raise R.T(R.Match(R.FS,fctname id,s,m),a)
                                              else raise exc 
        val _ = (ignore(pop()); pop_context())
        val sigma' = G.from_env c e'
    in  add_f id idl sigma sigma';
        (id,[(sid,sg',b)],se',a)
    end

  fun topdec td = case td of
      Sig(a,b) => I.Sig(sigdec a,b)
    | Fct(a,b) => I.Fct(fctdec a,b)
    | Str(a,b) => I.Str(strdec a,b)
    | SeqTopDec([l],b) => topdec l
    | SeqTopDec(l,b) => I.SeqTopDec(List.map topdec l,b) 

  (* return absyn and todo *)

  fun overload_resolve() = (
        List.iter (fn (t,a) => case !t of 
                                 Var(Ol _) => t := type_int;
                                              R.warning(R.Overload,a)
                               | _ => ()) (!Tv.Ol.m);
        Tv.Ol.m := [])

  fun infer p = let val _ = Scope.scope p
                    val p' = topdec p 
                in  overload_resolve();
                    Row.close_all();
                    p'
                end
end
