(*
 Oukseh Lee, The LET Project, KAIST

 Copyright(c) 2000-2002 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)


structure Labm = struct
  structure Label = struct type t = Info.StringInfo.labelinfo 
                           val compare = compare 
                    end
  structure M = OrderedListMapFn(Label)
  type key = Label.t
  type 'a t = 'a M.t

  exception InvalidIter2

  fun union s1 s2 = M.fold M.add s1 s2
  val length = List.length
  fun is_empty m = m = M.empty
  fun iter2 f [] [] = ()
    | iter2 f ((l1,t1)::o1) ((l2,t2)::o2) = (f l1 t1 l2 t2; iter2 f o1 o2)
    | iter2 f _ _ = raise InvalidIter2
  fun split m1 m2 = let
    fun f m k t (c,r) = if M.mem k m then (M.add k t c,r) else (c,M.add k t r)
    val (c1,rm1) = M.fold (f m2) m1 (M.empty,M.empty)
    val (c2,rm2) = M.fold (f m1) m2 (M.empty,M.empty)
   in
    ((c1,c2),(rm1,rm2))
   end   
  fun to_list m = m
  fun of_list l = List.fold_left (fn m (k,v) => M.add k v m) M.empty l
end 

structure Ty = struct
  open Info.StringInfo
  open String_ast.Ast
  open Format
  
  type ty = 
         Link of ty ref
       | Fun of ty * ty
       | Data of ty list * tyname
       | Tuple of ty list
       | Var of tyvar
       | Rec of ty Labm.t * (row ref) option
   and tyvar =
         Meta of int * int (* id / level *)
       | Ol of int (* 0 is bound *)
       | Name of int * string 
       | Bound of int
   and row =
         Rvar of int * int (* id / level *)
       | Row of ty Labm.t * (row ref) option

   and tyname = tyidinfo * int * int * string list
                (* name / identifier / arity / context *)
   and cons = (conidinfo * ty option) list

   
  type scheme = int * ty

  structure Tids = struct
    open Set
    structure M = Make(struct type t = tyidinfo val compare = Pervasives.compare end)
    fun singleton (x,_) = M.singleton x
    fun add (x,_) s = M.add x s
  end

  structure Tv = struct
    type t = ty
    val count = ref 0
    fun new() = (count++; Link(ref(Var(Meta(!count,!count)))))

    structure Ol = struct
      val m = ref ([] : (ty ref * Location.t) list)
      fun new a = (count++; let val v = ref(Var(Ol(!count))) 
                            in  m := (v,a)::(!m);
                                Link v
                            end)

    end
    fun new_named s = (count++; Var(Name((!count),s))) 
    fun current() = !count
    fun bound i = Var(Bound i)
    fun make_tvname n = 
      let fun f n = if n<26 then String.make 1 (Char.chr(97+n))
                            else (f (n/26))^(f (n % 26))
      in  f(n)
      end
  end

  structure Tn = struct
    type t = tyname
    val compare = compare
    val count = ref 0
    fun predefined s k = (s,0,k,[])
    fun new s k c = (s, count++; !count, k, c)
    fun current() = !count
    fun copy c (s,_,k,_) = new s k c
    fun arity (_,_,n,_) = n
    fun name (n,_,_,_) = n
    fun context (_,_,_,c) = c
  end 

  structure Tns = ListSetFn(Tn)

  fun list_init n f = let
       fun g m = if n=m then [] else (f m)::(g(m+1))
      in g 0 end
  fun init_tuple n f = Tuple(list_init n f)

  structure Row = struct
    val count = ref 1
    val m = ref []
    fun new_with_level i = (count++; 
          let val s = ref(Rvar(!count, i)) 
          in  count++;
              m := s::(!m);
              Some s 
          end)
    fun new() = new_with_level (!Tv.count)
    fun closerow m = Rec(m,None)
    fun openrow m = Rec(m,new())
    val openrow_from_olist = openrow
    fun flatten x = let
      fun f r = case !r of
            Rvar _ => (Labm.M.empty, Some r)
          | Row(m,None) => (m,None)
          | Row(m,Some r0) => let val (m1,r1) = f r0 in (Labm.union m m1,r1) end
     in
      case x of (m,None) => x
              | (m,Some r) => (case (!r) of Rvar _ => x
                                          | _ => f(ref(Row x)))
     end

    (* Convert a record to a tuple 
       tuplize: convert to a tuple if possible (an open record is not converted)
       enforce_tuple: convert to a tuple (an open record is closed) *)
    exception CannotTuplize
    fun labm2tuple m = init_tuple (Labm.length m) (fn i=>Labm.M.find (string_of_int i) m)
(*
    fun close r = case flatten r of (m,Some r) => (r := Row([],None); (m,None))
                                  | x => x 
    fun tuplize r = case flatten r of 
          (m,None) => ((labm2tuple m) handle _ => raise CannotTuplize)
        | _ => raise CannotTuplize
*)
    fun enforce_tuple r = case flatten r of
          (m,None) => ((labm2tuple m) handle _ => raise CannotTuplize)
        | (m,Some r) => (r := Row([],None); 
                         (labm2tuple m) handle _ => raise CannotTuplize)

    fun close_all () =
      let fun close r = case !r of
            Rvar _ => r := Row([],None)
          | Row(m,Some r') => close r'
          | _ => ()
      in  List.iter close (!m);
          m := []
      end

    fun of_tuple' l =
      let fun f _ [] = []
            | f n (h::t) = (string_of_int n, h)::(f (n+1) t)
      in  Labm.of_list(f 0 l)
      end
    fun of_tuple tl = Rec(of_tuple' tl,None)
  end

  val dummy = Tv.new
(*
  val static_dummy = Link(ref(Var(Meta(-1,-1))))
*)
  fun unlink x = case x of
        (Link t) => (case !t of Var(Meta _) | Var(Ol _) => x | x => unlink x)
      | _ => x
  
  fun gen_data x name = Data(x,name)
  fun new_data x name c = Data(x,Tn.new name (List.length x) c)

  fun gen_predef s= gen_data [] (Tn.predefined s 0)
  val type_int =    gen_predef "int"
  val type_real =   gen_predef "real"
  val type_string = gen_predef "string"
  val type_bool =   gen_predef "bool"
  val type_unit =   gen_predef "unit"
  val type_char =   gen_predef "char"
  val type_exn =    gen_predef "exn"

  fun get_predef t = case unlink t of Data([],(s,0,0,[])) => Some s | _ => None
  fun get_ref_predef t = case unlink t of Data([Data([],(s,0,0,[]))],
                                               ("ref",0,1,[])) => Some s | _ => None

  fun is_type_int t = get_predef t = Some "int"
  fun is_type_real t = get_predef t = Some "real"
  fun is_type_string t = get_predef t = Some "string"
  fun is_type_char t = get_predef t = Some "char"

  fun gen_predef t s = gen_data t (Tn.predefined s 1)
  fun gen_ref t =   gen_predef [t] "ref"

  fun gen_tuple l = Tuple l
  fun gen_pair t1 t2 = gen_tuple [t1,t2]
  fun gen_list t =  gen_predef [t] "list"
  fun gen_option t = gen_predef [t] "option" // for OCaml
  fun gen_array t = gen_predef [t] "array"
  fun gen_fun t1 t2 = Fun(t1,t2)

  fun is_fun t = case t of
        Fun _ | Var _ => true
      | Link t' => is_fun (!t')
	  | _ => false

  fun is_exn t = case unlink t of
        Data([],("exn",0,0,[]))
      | Fun(_,Data([],("exn",0,0,[]))) => true
      | _ => false

  fun exn_arg t = case unlink t of
        Fun(x,Data([],("exn",0,0,[]))) => Some x
      | _ => None

  val fold = List.fold_left



 
 open Hashtbl
 structure Intmap = Make(struct type t=int fun equal x y = x=y 
                                val hash = Hashtbl.hash end)

 structure Ungeneralized = struct
   val m = ref []
   fun add x = m := x::(!m)
   fun fix () = 
     let fun f t = case unlink t of
               Link t' => t' := type_unit
             | _ => ()
     in  List.iter f (!m);
         m := []
     end
   fun reset () = m := []
 end

 fun clos c t = let
(*   val _ = (printf "clos called (%d, " c; pp t; printf ").@.") *)
   val m = (Intmap.create 30): int Intmap.t
   val count = ref 0
   fun lookup_bound t = (Tv.bound(Intmap.find m t)) 
         handle Not_found => let val i = !count 
                             in count++; Intmap.add m t i; Tv.bound i end 
   fun f t = case t of
         Link t2 => (case !t2 of
           Var(Meta(x,y)) => if y<=c then t else lookup_bound x
         | Var(Ol _) => t
         | t => f t)
       | Fun(t1,t2) => Fun(f t1,f t2)
       | Data(ts,n) => Data(fs ts,n)
       | Tuple a => Tuple(fs a)
       | Rec r => Rec(case Row.flatten r of
           (m,x) => (Labm.M.map f m,x))
(*         | _ => raise TyDebug.Bug "Ty.clos.Row") *)
       | Var(Bound x) => lookup_bound (-x)
       | Var(Name(x,_)) => if x<=c then t else lookup_bound x
       | Var(Ol _ ) => t (* 2.001 *)
       | Var _ => raise TyDebug.Bug "Ty.bound.Var"
   and fs ts = List.map f ts
   val ft = f t
  in (!count,ft) end

 fun copy loc tvl t  = let
   val n = List.length tvl
   val m = Array.of_list tvl
   fun f t = case t of
         Link t2 => (case !t2 of
           Var(Meta _)
         | Var(Name _) => t
         | (t as Var(Ol 0)) => f t
         | Var(Ol _) => t
         | t => f t)
       | Fun(t1,t2) => Fun(f t1,f t2)
       | Data(ts,n) => Data(fs ts,n)
       | Tuple a => Tuple(fs a)
       | Rec r => Rec(case r of
           (m,None) => (Labm.M.map f m,None)
         | _ => (case Row.flatten r of (m,x) => (Labm.M.map f m,x)))
       | Var(Bound i) => if i>=0 andalso i<n then m.[i] 
                         else raise TyDebug.Bug "Ty.instance.Var(Bound _)"
       | Var(Ol 0) => (case m.[0] of 
                         Link t2 => ((case !t2 of
                           Var(Meta _) => m.[0] <- Tv.Ol.new loc
                         | _ => ()); m.[0])  
                      | _ => raise TyDebug.Bug "Ty.instance.Var(Ol _)")
       | _ => t
   and fs ts = List.map f ts
  in 
     (f t) handle _ => raise TyDebug.Bug "Ty.copy"
  end

 fun instance loc (n,t) = copy loc (list_init n (fn _ => Tv.new())) t
    
 fun in_type (_,t) = t
 fun in_scheme t = (0,t)

 fun used_tyids t = let
   fun f (ConstTy(tl,([],id,_),_)) = Tids.M.union (fs tl) (Tids.singleton id)
     | f (FunTy(t1,t2,_)) = Tids.M.union (f t1) (f t2)
     | f (RecordTy(l,_)) = fr l
     | f (TupleTy(l,_)) = fs l
     | f _ = Tids.M.empty
   and fs l = List.fold_right (fn t => Tids.M.union (f t)) l Tids.M.empty
   and fr l = List.fold_right (fn (_,t,_) => Tids.M.union (f t)) l Tids.M.empty
  in f t end

 structure Tvs = ListSetFn(struct type t = tyid 
                                  val compare = fn (x,_) (y,_) => compare x y end)
 fun used_tyvars t = let
   fun f (ConstTy(l,_,_)) = fs l
     | f (TupleTy(l,_)) = fs l
     | f (FunTy(t1,t2,_)) = Tvs.union (f t1) (f t2)
     | f (RecordTy(l,_)) = fr l
     | f (VarTy(id,_)) = Tvs.singleton id
   and fs l = List.fold_right (fn t => Tvs.union (f t)) l Tvs.empty
   and fr l = List.fold_right (fn (_,t,_) => Tvs.union (f t)) l Tvs.empty
  in f t end

 exception Not_equal 

 (* assume that input types satisfies the constraints after type inference *) 
 fun equal t1 t2 = 
   let fun f t1 t2 = case (t1,t2) of
         (Link t1',Link t2') => f (!t1') (!t2')
       | (Link t1',_) => f (!t1') t2
       | (_,Link t2') => f t1 (!t2')
       | (Fun(t11,t12), Fun(t21,t22)) => (f t11 t21; f t12 t22)
       | (Tuple l1, Tuple l2) => List.iter2 f l1 l2
       | (Tuple _,Rec r) => f t1 (Row.enforce_tuple r)
       | (Rec r,Tuple _) => f (Row.enforce_tuple r) t2
       | (Rec r1,Rec r2) => (case (Row.flatten r1,Row.flatten r2) of
                               ((m1,None),(m2,None)) => 
                                  Labm.iter2 (fn l1 t1 l2 t2 => 
                                                if l1=l2 then f t1 t2
                                                         else raise Not_equal) 
                                              m1 m2
                             | _ => raise TyDebug.Bug "Ty.equal.Rec")
       | (Data(l1,n1), Data(l2,n2)) => if n1=n2 then List.iter2 f l1 l2 
                                                else raise Not_equal
       | (x,y) => if x<>y then raise Not_equal
   in  (f t1 t2; true) handle (x as TyDebug.Bug _) => raise x | _ => false 
   end

  structure CE = struct
    type cons = (conidinfo * ty option) list
    structure M = MapFn(Tn)
    type t = cons M.t
    val listcons = let val a = Tv.bound 0
     in [("nil",None),("::",Some(gen_pair a (gen_list a)))]
     end
    val optioncons = let val a = Tv.bound 0
     in [("None",None),("Some",Some a)]
     end
    val boolcons = [("True",None),("False",None)]

    val initce = M.add (Tn.predefined "bool" 0) boolcons
                 (M.add (Tn.predefined "option" 1) optioncons
                 (M.add (Tn.predefined "list" 1) listcons M.empty))
    val m' = ref (initce:t)
    val m = ref (M.empty:t)
    fun init() = (m' := initce; m := M.empty)
    fun find tn = (M.find tn (!m)) handle Not_found => (
                  (M.find tn (!m')) handle Not_found => [])
    fun add tn [] = ()
      | add tn cons = m := M.add tn cons (!m)
    fun add_ce ce = M.iter (fn k v => m' := M.add k v (!m')) ce
    fun add_ce' ce = M.iter (fn k v => m := M.add k v (!m)) ce
(*
    fun print_all() = let
      fun pp_cons(a,None) = printf "%s@ " a 
        | pp_cons(a,Some t) = (printf "%s of "a; Print.typ' t; printf "@ ") 
      fun f k v = (Tn.pp k; printf " = @["; List.iter pp_cons v; printf "@]@.")
     in M.iter f (!m) end
*)
  end

  structure TS = struct
    type t = TsName of tyname (* data type *)
           | TsFun of scheme

    exception Apply_error

    fun arity ts = case ts of 
          TsName tn => Tn.arity tn
        | TsFun (n,_) => n

    fun normalize x = 
      let fun f i [] = true
            | f i (h::t) = if unlink h = Tv.bound i then f (i+1) t else false
      in  case x of
            TsFun(n,tau) => (case unlink tau of
              Data(tl,tn) => if Tn.arity tn = n andalso
                                List.length tl = n
                                andalso f 0 tl then TsName tn else x
            | _ => x)
          | _ => x
      end
 (*  
    fun gen_data (id,_) n = TsName(Tn.new id n)
*)
    fun gen_fun n tau = normalize(TsFun(n,tau))
    fun gen_name t = TsName t
    fun gen_var (id,_) n c = TsName(Tn.new id n c)

    fun tynames ts = case ts of
        TsName tn => Tns.singleton tn 
      | _ => Tns.empty

    fun apply tl ts = 
         if arity ts = List.length tl then 
          case ts of
            TsName tn => Data(tl,tn)
          | TsFun (n,tau) => copy Location.none tl tau
         else raise Apply_error

  end
end
